from django.db import models

from . import forms
from .validators import AntiVirusValidator, FileContentTypeValidator, \
    FileExtensionValidator


class SafeFileField(models.FileField):
    def __init__(self, **kwargs):
        self.allowed_extensions = kwargs.pop('allowed_extensions', None)
        self.check_content_type = kwargs.pop('check_content_type', True)
        self.scan_viruses = kwargs.pop('scan_viruses', False)

        default_validators = []

        if self.allowed_extensions:
            default_validators.append(
                FileExtensionValidator(self.allowed_extensions)
            )

        if self.check_content_type:
            default_validators.append(FileContentTypeValidator())

        if self.scan_viruses:
            default_validators.append(AntiVirusValidator())

        self.default_validators = default_validators + self.default_validators

        super().__init__(**kwargs)

    def formfield(self, **kwargs):
        return super().formfield(
            form_class=forms.SafeFileField,

            allowed_extensions=self.allowed_extensions,
            check_content_type=self.check_content_type
        )



from urllib.parse import urlparse

import clamd
from django.conf import settings
from django.utils.functional import SimpleLazyObject

from safe_filefield import default_settings


def get_scanner(socket, timeout=None):
    if socket.startswith('unix://'):
        return clamd.ClamdUnixSocket(socket[7:], timeout)
    elif socket.startswith('tcp://'):
        uri = urlparse(socket)

        return clamd.ClamdNetworkSocket(
            uri.hostname, uri.port or 3310, timeout
        )
    else:
        raise NotImplementedError(
            'Missed or unsupported ClamAV connection string schema. '
            'Only tcp:// or unix:// is allowed.'
        )


def _get_default_scanner():
    return get_scanner(
        getattr(settings, 'CLAMAV_SOCKET', default_settings.CLAMAV_SOCKET),
        getattr(settings, 'CLAMAV_TIMEOUT', default_settings.CLAMAV_TIMEOUT),
    )


scanner = SimpleLazyObject(_get_default_scanner)


def scan_file(f):
    _pos = f.tell()
    f.seek(0)
    status, virus_name = scanner.instream(f)['stream']
    f.seek(_pos)

    return status, virus_name



from django import forms

from .validators import AntiVirusValidator, FileContentTypeValidator, \
    FileExtensionValidator


class SafeFileField(forms.FileField):
    def __init__(self, **kwargs):
        self.allowed_extensions = kwargs.pop('allowed_extensions', None)
        self.check_content_type = kwargs.pop('check_content_type', True)
        self.scan_viruses = kwargs.pop('scan_viruses', False)

        default_validators = []

        if self.allowed_extensions:
            default_validators.append(
                FileExtensionValidator(self.allowed_extensions)
            )

        if self.check_content_type:
            default_validators.append(FileContentTypeValidator())

        if self.scan_viruses:
            default_validators.append(AntiVirusValidator())

        self.default_validators = default_validators + self.default_validators

        super().__init__(**kwargs)



from setuptools import find_packages, setup

setup(
    name='django-safe-filefield',
    version='1.0.0',
    url='https://github.com/mixkorshun/django-safe-filefield',
    description='Secure file field, which allows you to '
                'restrict uploaded file extensions.',
    keywords=['django', 'filefield', 'model-field', 'form-field'],

    long_description=open('README.rst', 'r').read(),
    long_description_content_type="text/x-rst",

    author='Vladislav Bakin',
    author_email='mixkorshun@gmail.com',
    maintainer='Vladislav Bakin',
    maintainer_email='mixkorshun@gmail.com',

    license='MIT',

    install_requires=[
        'django',

        'python-magic',
        'clamd',
    ],

    packages=find_packages(exclude=['*.tests.*', '*.tests']),

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Security',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',

        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4',
        'Framework :: Django :: 4.2',
    ],
)



import mimetypes
import os

from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from . import clamav
from .utils import detect_content_type


@deconstructible
class FileExtensionValidator(object):
    message = _(
        "File extension '%(extension)s' is not allowed. "
        "Allowed extensions are: '%(allowed_extensions)s'."
    )
    code = 'invalid_extension'

    def __init__(self, allowed_extensions=None, message=None, code=None):
        self.allowed_extensions = allowed_extensions

        if message is not None:
            self.message = message

        if code is not None:
            self.code = code

    def __call__(self, value):
        extension = os.path.splitext(value.name)[1][1:].lower()

        if (self.allowed_extensions is not None) \
                and (extension not in self.allowed_extensions):
            raise ValidationError(
                self.message,
                code=self.code,
                params={
                    'extension': extension,
                    'allowed_extensions': ', '.join(
                        self.allowed_extensions)
                }
            )


class FileContentTypeValidator:
    message = _(
        'File has invalid content-type. '
        'Maybe file extension is equal to file content?'
    )

    code = 'invalid_content_type'

    def __init__(self, message=None, code=None):
        if message is not None:
            self.message = message

        if code is not None:
            self.code = code

    def __call__(self, file):
        if hasattr(file, '_get_file'):
            file = file._get_file()

        __, ext = os.path.splitext(file.name)

        detected_content_type = detect_content_type(file)

        if getattr(file, 'content_type', None) is not None:
            is_valid_content_type = bool(
                (ext in mimetypes.guess_all_extensions(detected_content_type)
                 and ext in mimetypes.guess_all_extensions(file.content_type)
                 ) or (detected_content_type == 'application/CDFV2-unknown'
                       and file.content_type == mimetypes.guess_type('.doc')
                       and ext == "doc")
            )
            params = {
                'extension': ext,
                'content_type': file.content_type,
                'detected_content_type': detected_content_type
            }
        else:
            is_valid_content_type = bool(
                (ext in mimetypes.guess_all_extensions(detected_content_type))
                or (detected_content_type == 'application/CDFV2-unknown'
                    and ext == "doc")
            )
            params = {
                'extension': ext,
                'content_type': None,
                'detected_content_type': detected_content_type
            }

        if not is_valid_content_type:
            raise ValidationError(
                self.message,
                code=self.code,
                params=params
            )


class AntiVirusValidator:
    message = _('File is infected with %(virus)s.')

    code = 'infected'

    def __init__(self, message=None, code=None):
        if message is not None:
            self.message = message

        if code is not None:
            self.code = code

    def __call__(self, file):
        status, virus_name = clamav.scan_file(file)

        if status != 'OK':
            raise ValidationError(
                self.message,
                self.code,
                params={
                    'virus': virus_name
                }
            )
